/*
     MDIApp.CC
     Author: Michael Nuwer
     Date..: February 2002 

     DiveShop Tutorial project -- Global Application class

     
*/
class DiveShopMDIApp               

   this.FrameWinText = "DiveShop Tutorial Project"

   // Set procedures ...:
   set procedure to START.PRG    additive
   set procedure to DiveMenu.mnu additive
   set procedure to toolbars.cc  additive
   set procedure to CustButt.cc  additive
   set procedure to FormCntl.cc  additive
   set procedure to Orders.wfm   additive

   //  Create a referance for this object and 
   //  attach it to the built-in "_app" object.
   _app.myApp = this

   function open
      // set a reference to the menu
      this.rootMenu = new DiveMenu(_app.framewin,"Root")

      // Make sure no forms are open
      close forms

      // Make sure we're not in "design mode"
      set design off

      // set the <Escape> key "off" but store the 
      // current setting:
      this.OldEscape = set("ESCAPE")
      set escape off

      // Turn off the Visual dBASE shell, but 
      // leave the MDI frame window:
      shell( false, true )

      // Turn off the application's speedBar and statusBar
      _app.speedbar  := false
      _app.statusBar := false

      // Set the text property for the application's framewin
      _app.framewin.oldText = _app.framewin.text
      _app.framewin.text    := this.FrameWinText

      ///////////////////////////////////////////////////
      // create an instance of Toobar
      ///////////////////////////////////////////////////
      this.toolbar = new Toolbar1()
      this.toolbar.attach(_app.framewin)

      ///////////////////////////////////////////////////
      // Prototype API funtions used for "Tile Window"
      // and "Cascade Window" in the menu.
      ///////////////////////////////////////////////////
      if type( "GetParent" ) # "FP"
        extern CHANDLE GetParent(CHANDLE) User32
      endif

      if type( "SendMessage" ) # "FP"
         extern CLONG SendMessage(CHANDLE,CUINT,CWORD,CLONG ) ;
                User32 from "SendMessageA"
      endif

      ///////////////////////////////////////////////////
      //  prototype the winHelp function
      //  See WinHelp by Gery White for details
      //////////////////////////////////////////////////
      if type("WinHelp") # "FP"
         extern clogical WinHelp(chandle, cstring, cuint, cstring) user32 ;
            from "WinHelpA"
      endif

      // set the help file
      this.helpFile = set('dire')+"\DiveShop.hlp"

      //////////////////////////////////////////////////////
      // make sure application is "on top"
      ///////////////////////////////////////////////////
      if type('BringWindowToTop') # 'FP'
      	extern clogical BringWindowToTop( cHandle ) user32
      endif
      BringWindowToTop( _app.framewin.hWnd )


   function close

      // close any forms that might have been left open
      close forms

      //////////////////////////////////////////////////
      // release toolbar
      ///////////////////////////////////////////////////
      t=this.toolbar
      t.detach(_app.framewin)
      release object t
      t := null

      //////////////////////////////////////////////////
      // if we are in the "runtime" environment,
      // we want to "quit" (otherwise the framewin will
      // be left on screen)
      if ( "runtime" $ lower( version(0) ) )
         quit
      else
         // otherwise, let's reset some values:
         with ( _app )
            myApp  := null
            framewin.text := framewin.OldText
            speedBar      := true
            statusBar     := true
         endwith

         // go back to design mode:
         set design on
         
	      // set escape back to whatever it's previous
         // state was:
	      cEscape = this.oldEscape
         set escape &cEscape.

	      // close any open procedures
         set procedure to

         // release the menu
         _app.framewin.root.release()

         // set the shell back ...
         shell( true, true )

      endif



   function openForm( oForm )
     // set the top/left properties so the form isn't
     // *right* on top of the current one ...
     if type( "this.currentForm" ) == "O"
        oForm.top  := this.currentForm.top + 2
        oForm.left := this.currentForm.left + 2
     else
        oForm.top  := 0
        oForm.left := 10
     endif

     oForm.open()
     oForm.setFocus()
   return


   ///////////////////////////////////////////////////
   ///
   ///   Methods used to open specific forms 
   ///   These methods are called from the Menu or
   ///   from the toolbar
   ///
   ///////////////////////////////////////////////////
   function openOrderForm
      local f
      f = new OrdersForm()
      this.openForm( f )
      return

   function openCustListForm
      local f
      set procedure to custlist.wfm addi
      f = new custListForm()
      f.text = 'Find a Customer'
      f.mode = 'non-modal'
      this.openForm( f )
      return

   function openStockListForm
      local f
      set procedure to stocklist.wfm addi
      f = new stockListForm()
      f.text = 'Find a Stock Item'
      f.mode = 'non-modal'
      this.openForm( f )
      return

   function openVendorListForm
      local f
      set procedure to vendorlist.wfm addi
      f = new vendorListForm()
      f.text = 'Find a Vendor'
      f.mode = 'non-modal'
      this.openForm( f )
      return

   function closeForm
      // close current window
      if type( "_app.myApp.currentForm" ) # "U" AND;
         type( "_app.myApp.currentForm" ) # "NULL"
         _app.myApp.currentForm.close()
      else
         msgbox( "The currently active form cannot be closed with this option.",;
                 "Can't do it!", 64 )
      endif
      return


   ///////////////////////////////////////////////////
   ///
   ///   Methods used to navigate the currentForm
   ///   rowset and to cut, copy, and paste 
   ///   from or to the activeControl.
   ///   These methods are called from the Menu or
   ///   from the toolbar
   ///
   ///////////////////////////////////////////////////

   function firstRow
      this.currentForm.rowset.first()
   	return

   function prevRow
      this.currentForm.rowset.next(-1)
      return

   function nextRow
      this.currentForm.rowset.next()
   	return

   function lastRow
      this.currentForm.rowset.last()
   	return

   function cutText
   	if type('this.currentForm.activeControl.cut') # 'U'
      	this.currentForm.activeControl.cut()
      endif
   	return

   function copyText
   	if type('this.currentForm.activeControl.copy') # 'U'
      	this.currentForm.activeControl.copy()
      endif
   	return

   function pasteText
   	if type('this.currentForm.activeControl.paste') # 'U'
      	this.currentForm.activeControl.paste()
      endif
   	return


endclass // don't forget this!
