// navigation buttons:
#define TOPTIP                  "First row in rowset"
#define PREVTIP                 "Previous row in rowset"
#define NEXTTIP                 "Next row in rowset"
#define ENDTIP                  "Last row in rowset"
// clipboard buttons:
#define CUTTIP                  "Cut selected text"
#define COPYTIP                 "Copy selected text"
#define PASTETTIP               "Paste from clipboard"

#define EXITTIP                 "Exit the program"
#define ORDERSTIP               "Open the Orders form"
#define CUSTTIP                 "Open the Customer list"
#define STOCKTIP                "Open the Stock list"
#define VENDORTIP               "Open the Vendors list"
#define CLOSETIP                "Close the current form"

#DEFINE ToolButtonWidth 17    // for standard "small" toolbutton
                              // images
#DEFINE BitMapWidth2    20    // for larger images, so they don't
                              // get cut off

/*
   --------------------------------------------------------
   ToolBars.CC 

   Adapted from toolbar.cc by Ken Mayer (found in dUFLP)

   */

class ToolBar1 of Toolbar
   with (this)
      floating := false
   endwith
   this.canFloat = false

   this.tbOrders1    = new tbOrders(this)
   this.tbCust1      = new tbCust(this)
   this.tbStock1     = new tbStock(this)
   this.tbVendors1   = new tbVendors(this)

   this.tbSep1      = new tbSeparator(this)

   this.closeForm1    = new closeForm(this)
   // separator:
   this.tbSep2      = new tbSeparator(this)

   // Clipboard buttons:
   this.tbCut1      = new tbCut(this)
   this.tbCopy1     = new tbCopy(this)
   this.tbPaste1    = new tbPaste(this)

   // separator:
   this.tbSep3      = new tbSeparator(this)

   // Navigation buttons:
   this.tbTop1      = new tbTop(this)
   this.tbPrevious1 = new tbPrevious(this)
   this.tbNext1     = new tbNext(this)
   this.tbEnd1      = new tbEnd(this)


   this.tbSep4      = new tbSeparator(this)
   this.tbExit1 = new tbExit(this)

   function onUpdate // when toolbar updates
      if this.canFloat == false and this.floating
         this.floating := false
      endif
      // is there a rowset?
      lRowset = type('_app.myApp.currentForm.rowset') == 'O'

      // if we do have a rowset, check to see if we're at the
      // first/last records:
      if lRowset     
         // Navigate buttons
         lAtFirst = _app.myApp.currentForm.rowset.atFirst()
         lAtLast  = _app.myApp.currentForm.rowset.atLast()
         this.tbTop1.enabled      := NOT lAtFirst
         this.tbPrevious1.enabled := NOT lAtFirst

         this.tbNext1.enabled     := NOT lAtLast
         this.tbEnd1.enabled      := NOT lAtLast
      else
         // no rowset
         this.tbTop1.enabled      := false
         this.tbPrevious1.enabled := false
         this.tbNext1.enabled     := false
         this.tbEnd1.enabled      := false
      endif

      // Clipboard buttons:
      lClip = ( TYPE("_app.myApp.currentform.activeControl.copy") == "FP" )
      this.tbCut1.enabled   := lClip
      this.tbCopy1.enabled  := lClip
      this.tbPaste1.enabled := lClip


endclass

/* =======================================================
   Now for the Button definitions themselves:
   ======================================================= */

/* =======================================================
   A Base Class: 
   ======================================================= */
class BaseToolButton( parentObj ) of ToolButton( parentObj ) custom
   this.twoState := false
   this.rowset   =  NULL // custom property!
   this.bitmapwidth := ToolButtonWidth
endclass

/* =======================================================
   A separator button
   ======================================================= */
class tbSeparator( parentObj ) of ToolButton( parentObj ) custom
    this.separator   := true
    this.bitmapwidth := 2
endclass

/* =======================================================
   The subclassed buttons:
   ======================================================= */

/* ---------------------------------------------------- */
class tbTop( parentObj ) of BaseToolButton( parentObj ) custom

   this.BitMap   := "RESOURCE:2 TS_FIRST"
   this.SpeedTip := TOPTIP

   function onClick
      _app.myApp.firstRow()
     return
endclass


class tbPrevious( parentObj ) of BaseToolButton( parentObj ) custom

   this.BitMap   := "RESOURCE:2 TS_PREV"
   this.SpeedTip := PREVTIP

   function onClick
      _app.myApp.prevRow()
      return
endclass


class tbNext( parentObj ) of BaseToolButton( parentObj ) custom

   this.BitMap   := "RESOURCE:2 TS_NEXT"
   this.SpeedTip := NEXTTIP

   function onClick
      _app.myApp.nextRow()
endclass


class tbEnd( parentObj ) of BaseToolButton( parentObj ) custom

   this.BitMap   := "RESOURCE:2 TS_LAST"
   this.SpeedTip := ENDTIP

   function onClick
      _app.myApp.LastRow()
      return
endclass

class tbCut( parentObj ) of BaseToolButton( parentObj ) custom

   this.Bitmap         := "RESOURCE:2 TS_CUT"
   this.SpeedTip       := CUTTIP

   function onClick
      _app.myApp.cutText()
      return
endclass

class tbCopy( parentObj ) of BaseToolButton( parentObj ) custom

   this.Bitmap         := "RESOURCE:2 TS_COPY"
   this.SpeedTip       := COPYTIP

   function onClick
      _app.myApp.copyText()
      return

endclass

class tbPaste( parentObj ) of BaseToolButton( parentObj ) custom

   this.Bitmap         := "RESOURCE:2 TS_PASTE"
   this.SpeedTip       := PASTETTIP

   function onClick
      _app.myApp.pasteText()
      return

endclass


class tbExit( parentObj ) of BaseToolButton( parentObj ) custom

   this.Bitmap         := "RESOURCE #28"
   this.SpeedTip       := EXITTIP

   function onClick
      _app.myApp.close()
      return

endclass

class tbOrders( parentObj ) of BaseToolButton( parentObj ) custom

   this.Bitmap         := "FILENAME openOrders.bmp"
   this.SpeedTip       := ORDERSTIP

   function onClick
      _app.myApp.openOrderForm()
      return
endclass


class tbCust( parentObj ) of BaseToolButton( parentObj ) custom

   this.Bitmap         := "FILENAME openCust.bmp"
   this.SpeedTip       := CUSTTIP 

   function onClick
      _app.myApp.openCustListForm()
      return
endclass


class tbStock( parentObj ) of BaseToolButton( parentObj ) custom

   this.Bitmap         := "FILENAME openStock.bmp"
   this.SpeedTip       := STOCKTIP 

   function onClick
      _app.myApp.openStockListForm()
      return
endclass


class tbVendors( parentObj ) of BaseToolButton( parentObj ) custom

   this.Bitmap         := "FILENAME openVendors.bmp"
   this.SpeedTip       := VENDORTIP 

   function onClick
      _app.myApp.openVendorListForm()
      return
endclass


class closeForm( parentObj ) of BaseToolButton( parentObj ) custom

   this.Bitmap         := "FILENAME closeForm.bmp"
   this.SpeedTip       := CLOSETIP 

   function onClick
      _app.myApp.closeForm()
      return
endclass


